#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import unittest

import fixtures  # type: ignore

from autopage.tests import sinks

import autopage


class LineBufferFromInputTest(unittest.TestCase):
    def test_tty(self) -> None:
        with sinks.TTYFixture() as inp:
            self.assertTrue(autopage.line_buffer_from_input(inp.stream))

    def test_file(self) -> None:
        with sinks.TempFixture() as inp:
            self.assertFalse(autopage.line_buffer_from_input(inp.stream))

    def test_default_tty(self) -> None:
        with sinks.TTYFixture() as inp:
            with fixtures.MonkeyPatch('sys.stdin', inp.stream):
                self.assertTrue(autopage.line_buffer_from_input())

    def test_default_file(self) -> None:
        with sinks.TempFixture() as inp:
            with fixtures.MonkeyPatch('sys.stdin', inp.stream):
                self.assertFalse(autopage.line_buffer_from_input())



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse

import fixtures  # type: ignore

import typing

import autopage
import autopage.argparse


class ArgumentParseTest(fixtures.TestWithFixtures):
    def setUp(self) -> None:
        self.useFixture(fixtures.EnvironmentVariable('PYTHON_COLORS', '1'))
        patch_ap = self.useFixture(fixtures.MockPatch('autopage.AutoPager'))
        self.pager = patch_ap.mock
        self.stream = self.useFixture(fixtures.StringStream('stdout')).stream
        self.pager.return_value.__enter__.return_value = self.stream

    def test_argparse(self,
                      module: typing.Any = autopage.argparse,
                      color: bool = True) -> None:
        self.pager.return_value.to_terminal.return_value = color
        parser = module.ArgumentParser()
        try:
            parser.parse_args(['foo', '--help'])
        except SystemExit as exit:
            self.assertIs(self.pager.return_value.exit_code.return_value,
                          exit.code)
        self.pager.assert_called_once_with(None,
                                           allow_color=True,
                                           line_buffering=False,
                                           reset_on_exit=False)
        self.pager.return_value.__enter__.assert_called_once()
        self.stream.seek(0)
        output = self.stream.read()
        self.assertEqual(any(c in output for c in ('\033', '\x1b')), color)

    def test_argparse_no_color(self) -> None:
        self.test_argparse(color=False)

    def test_monkey_patch(self, color: bool = True) -> None:
        patch = self.useFixture(fixtures.MockPatch('argparse._HelpAction'))
        autopage.argparse.monkey_patch()
        self.assertIsNot(patch.mock, argparse._HelpAction)
        self.test_argparse(argparse, color)

    def test_monkey_patch_no_color(self) -> None:
        self.test_monkey_patch(color=False)

    def test_monkey_patch_context(self, color: bool = True) -> None:
        patch = self.useFixture(fixtures.MockPatch('argparse._HelpAction'))
        with autopage.argparse.monkey_patch():
            self.assertIsNot(patch.mock, argparse._HelpAction)
            self.test_argparse(argparse, color)
        self.assertIs(patch.mock, argparse._HelpAction)

    def test_monkey_patch_context_no_color(self) -> None:
        self.test_monkey_patch_context(color=False)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import itertools
import unittest
import sys

from autopage.tests import isolation

import typing

import autopage


MAX_LINES_PER_PAGE = isolation.LINES - 1


class finite:
    def __init__(self, num_lines: int, **kwargs: typing.Any):
        self.num_lines = num_lines
        self.kwargs = kwargs

    def __call__(self) -> int:
        ap = autopage.AutoPager(pager_command=autopage.command.Less(),
                                **self.kwargs)
        with ap as out:
            for i in range(self.num_lines):
                print(i, file=out)
        return ap.exit_code()


def infinite() -> int:
    ap = autopage.AutoPager(pager_command=autopage.command.Less())
    try:
        with ap as out:
            for i in itertools.count():
                print(i, file=out)
    except KeyboardInterrupt:
        pass
    return ap.exit_code()


def from_stdin() -> int:
    ap = autopage.AutoPager(pager_command=autopage.command.Less(),
                            line_buffering=autopage.line_buffer_from_input())
    with ap as out:
        try:
            for line in sys.stdin:
                print(line.rstrip(), file=out)
        except KeyboardInterrupt:
            pass
    return ap.exit_code()


def with_exception() -> int:
    class MyException(Exception):
        pass

    ap = autopage.AutoPager(pager_command=autopage.command.Less())
    try:
        with ap as out:
            for i in range(50):
                print(i, file=out)
            raise MyException()
    except MyException:
        pass
    return ap.exit_code()


def with_stderr_output() -> int:
    ap = autopage.AutoPager(pager_command=autopage.command.Less())
    with ap as out:
        for i in range(50):
            print(i, file=out)
    print("Hello world", file=sys.stderr)
    return ap.exit_code()


class InvokePagerTest(unittest.TestCase):
    def test_page_to_end(self) -> None:
        num_lines = 100
        with isolation.isolate(finite(num_lines)) as env:
            pager = isolation.PagerControl(env)

            lines = num_lines
            while lines > 0:
                expected = min(lines, MAX_LINES_PER_PAGE)
                self.assertEqual(expected, pager.advance())
                lines -= expected
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.quit())

            self.assertEqual(num_lines, pager.total_lines())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_page_to_middle(self) -> None:
        num_lines = 100
        with isolation.isolate(finite(num_lines)) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())
            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())
            self.assertEqual(MAX_LINES_PER_PAGE, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_exit_pager_early(self) -> None:
        with isolation.isolate(infinite) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())
            self.assertEqual(MAX_LINES_PER_PAGE, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(141, env.exit_code())

    def test_interrupt_early(self) -> None:
        with isolation.isolate(infinite) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())
            env.interrupt()
            while pager.advance():
                continue
            pager.quit()
            self.assertGreater(pager.total_lines(), MAX_LINES_PER_PAGE)
            self.assertFalse(env.error_output())
        self.assertEqual(130, env.exit_code())

    def test_interrupt_early_quit(self) -> None:
        with isolation.isolate(infinite) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())
            env.interrupt()
            pager.quit()
            self.assertGreater(pager.total_lines(), MAX_LINES_PER_PAGE)
            self.assertFalse(env.error_output())
        self.assertEqual(130, env.exit_code())

    def test_interrupt_in_middle_after_complete(self) -> None:
        num_lines = 100
        with isolation.isolate(finite(num_lines)) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(MAX_LINES_PER_PAGE, pager.advance())

            for i in range(100):
                env.interrupt()

            self.assertEqual(MAX_LINES_PER_PAGE, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_interrupt_at_end_after_complete(self) -> None:
        num_lines = 100
        with isolation.isolate(finite(num_lines)) as env:
            pager = isolation.PagerControl(env)

            while pager.advance():
                continue

            self.assertEqual(num_lines, pager.total_lines())

            for i in range(100):
                env.interrupt()

            self.assertEqual(0, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_short_output(self) -> None:
        num_lines = 10
        with isolation.isolate(finite(num_lines)) as env:
            pager = isolation.PagerControl(env)

            for i, l in enumerate(pager.read_lines(num_lines)):
                self.assertEqual(str(i), l.rstrip())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_short_output_reset(self) -> None:
        num_lines = 10
        with isolation.isolate(finite(num_lines, reset_on_exit=True)) as env:
            pager = isolation.PagerControl(env)

            self.assertEqual(num_lines, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_short_streaming_output(self) -> None:
        num_lines = 10
        with isolation.isolate(from_stdin, stdin_pipe=True) as env:
            pager = isolation.PagerControl(env)

            with env.stdin_pipe() as in_pipe:
                for i in range(num_lines):
                    print(i, file=in_pipe)

            for i, l in enumerate(pager.read_lines(num_lines)):
                self.assertEqual(i, int(l))

            env.interrupt()
            self.assertEqual(0, pager.quit())
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_exception(self) -> None:
        num_lines = 50
        with isolation.isolate(with_exception) as env:
            pager = isolation.PagerControl(env)

            lines = num_lines
            while lines > 0:
                expected = min(lines, MAX_LINES_PER_PAGE)
                self.assertEqual(expected, pager.advance())
                lines -= expected
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.quit())

            self.assertEqual(num_lines, pager.total_lines())
            self.assertFalse(env.error_output())
        self.assertEqual(1, env.exit_code())

    def test_stderr_output(self) -> None:
        num_lines = 50
        with isolation.isolate(with_stderr_output) as env:
            pager = isolation.PagerControl(env)

            lines = num_lines
            while lines > 0:
                expected = min(lines, MAX_LINES_PER_PAGE)
                self.assertEqual(expected, pager.advance())
                lines -= expected
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.advance())
            self.assertEqual(0, pager.quit())

            self.assertEqual(num_lines, pager.total_lines())
            self.assertEqual('Hello world\n', env.error_output())
        self.assertEqual(0, env.exit_code())


class NoPagerTest(unittest.TestCase):
    def test_pipe_output_to_end(self) -> None:
        num_lines = 100
        with isolation.isolate(finite(num_lines),
                               stdout_pipe=True) as env:
            with env.stdout_pipe() as out:
                output = out.readlines()

            self.assertEqual(num_lines, len(output))
            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())

    def test_exit_early(self) -> None:
        with isolation.isolate(infinite, stdout_pipe=True) as env:
            with env.stdout_pipe() as out:
                for i in range(500):
                    out.readline()
            self.assertFalse(env.error_output())
        self.assertEqual(141, env.exit_code())

    def test_exit_early_buffered(self) -> None:
        num_lines = 10
        with isolation.isolate(from_stdin,
                               stdin_pipe=True, stdout_pipe=True) as env:
            with env.stdin_pipe() as in_pipe:
                for i in range(num_lines):
                    print(i, file=in_pipe)
            with env.stdout_pipe():
                # Close output without reading contents of buffer
                pass
            self.assertFalse(env.error_output())
        self.assertEqual(141, env.exit_code())

    def test_interrupt_early(self) -> None:
        with isolation.isolate(infinite, stdout_pipe=True) as env:
            env.interrupt()
            with env.stdout_pipe() as out:
                output = out.readlines()
            self.assertGreater(len(output), 0)
            self.assertFalse(env.error_output())
        self.assertEqual(130, env.exit_code())

    def test_short_streaming_output(self) -> None:
        num_lines = 10
        with isolation.isolate(from_stdin,
                               stdin_pipe=True, stdout_pipe=True) as env:
            with env.stdin_pipe() as in_pipe:
                for i in range(num_lines):
                    print(i, file=in_pipe)

            with env.stdout_pipe() as out:
                for i in range(num_lines):
                    self.assertEqual(i, int(out.readline()))
                env.interrupt()
                self.assertEqual(0, len(out.readlines()))

            self.assertFalse(env.error_output())
        self.assertEqual(0, env.exit_code())



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
This module provides a drop-in replacement for the standard library argparse
module, with the output of the 'help' action automatically sent to a pager
when appropriate.

To use, replace the code:

    >>> import argparse

with:

    >>> from autopage import argparse

Or, alternatively, call the ``autopage.argparse.monkey_patch()`` function to
monkey-patch the argparse module. This is useful when you do not control the
code that creates the ArgumentParser. The result of calling this function can
also be used as a context manager to ensure that the original functionality is
restored.
"""

import argparse
import contextlib
import functools
import os
import sys
import types
from typing import Any, Sequence, Text, TextIO, Tuple, Type, Optional, Union
from typing import Callable, ContextManager, Generator

import autopage

from argparse import *  # noqa


_HelpFormatter = argparse.HelpFormatter

_color_attr = '_autopage_color'
_original_stream_attr = '_autopage_original_stream'


def help_pager(out_stream: Optional[TextIO] = None) -> autopage.AutoPager:
    """Return an AutoPager suitable for help output."""
    return autopage.AutoPager(out_stream,
                              allow_color=True,
                              line_buffering=False,
                              reset_on_exit=False)


def use_color_for_parser(parser: argparse.ArgumentParser,
                         color: bool) -> None:
    """Configure a parser whether to output in color from HelpFormatters."""
    if os.getenv('FORCE_COLOR') is None:
        if (os.getenv('NO_COLOR') is not None or
            os.getenv('TERM') == 'dumb' or
            (not sys.flags.ignore_environment and
             os.getenv('PYTHON_COLORS') == '0')):
            color = False
        elif sys.platform == 'win32':
            try:
                import nt
                if not nt._supports_virtual_terminal():
                    color = False
            except (ImportError, AttributeError):
                color = False
    parser.color = color


class ColorHelpFormatter(_HelpFormatter):
    class _Section(_HelpFormatter._Section):  # type: ignore
        @property
        def heading(self) -> Optional[Text]:
            if (not self._heading
                    or self._heading == argparse.SUPPRESS
                    or not getattr(self.formatter, _color_attr, False)):
                return self._heading
            return f'\033[4m{self._heading}\033[0m'

        @heading.setter
        def heading(self, heading: Optional[Text]) -> None:
            self._heading = heading

    def _metavar_formatter(self,
                           action: argparse.Action,
                           default_metavar: Text) -> Callable[[int],
                                                              Tuple[str, ...]]:
        get_metavars = super()._metavar_formatter(action, default_metavar)
        if not getattr(self, _color_attr, False):
            return get_metavars

        def color_metavar(size: int) -> Tuple[str, ...]:
            return tuple(f'\033[3m{mv}\033[0m' for mv in get_metavars(size))
        return color_metavar


class ColorRawDescriptionHelpFormatter(ColorHelpFormatter,
                                       argparse.RawDescriptionHelpFormatter):
    """Help message formatter which retains any formatting in descriptions."""


class ColorRawTextHelpFormatter(ColorHelpFormatter,
                                argparse.RawTextHelpFormatter):
    """Help message formatter which retains formatting of all help text."""


class ColorArgDefaultsHelpFormatter(ColorHelpFormatter,
                                    argparse.ArgumentDefaultsHelpFormatter):
    """Help message formatter which adds default values to argument help."""


class ColorMetavarTypeHelpFormatter(ColorHelpFormatter,
                                    argparse.MetavarTypeHelpFormatter):
    """Help message formatter which uses the argument 'type' as the default
    metavar value (instead of the argument 'dest')"""


class _HelpAction(argparse._HelpAction):
    def __init__(self,
                 option_strings: Sequence[Text],
                 dest: Text = argparse.SUPPRESS,
                 default: Text = argparse.SUPPRESS,
                 help: Optional[Text] = None) -> None:
        argparse.Action.__init__(
            self,
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Union[Text, Sequence[Any], None],
                 option_string: Optional[Text] = None) -> None:
        pager = help_pager()
        with pager as out:
            if out is not pager._out:
                setattr(out, _original_stream_attr, pager._out)
            use_color_for_parser(parser, (pager.to_terminal() and
                                          getattr(parser, 'color', True)))
            parser.print_help(out)
        parser.exit(pager.exit_code())


class _ActionsContainer(argparse._ActionsContainer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.register('action', 'help', _HelpAction)


def _substitute_formatter(
            get_fmtr: Callable[[Any], _HelpFormatter]
        ) -> Callable[[argparse.ArgumentParser], _HelpFormatter]:
    @functools.wraps(get_fmtr)
    def _get_formatter(parser: argparse.ArgumentParser,
                       file: Optional[TextIO] = None) -> _HelpFormatter:
        if parser.formatter_class is _HelpFormatter:
            parser.formatter_class = HelpFormatter
        kwargs = {}
        if file is not None:
            kwargs['file'] = getattr(file, _original_stream_attr, file)
        formatter = get_fmtr(parser, **kwargs)
        if isinstance(formatter, ColorHelpFormatter):
            setattr(formatter, _color_attr,
                    getattr(parser, 'color', False))
        return formatter
    return _get_formatter


class AutoPageArgumentParser(argparse.ArgumentParser, _ActionsContainer):
    @_substitute_formatter
    def _get_formatter(self, file: Optional[TextIO] = None) -> _HelpFormatter:
        kwargs = {}
        if file is not None:
            kwargs['file'] = file
        return super()._get_formatter(**kwargs)


ArgumentParser = AutoPageArgumentParser   # type: ignore
if sys.version_info < (3, 14):
    HelpFormatter = ColorHelpFormatter    # type: ignore
    RawDescriptionHelpFormatter = \
        ColorRawDescriptionHelpFormatter  # type: ignore
    RawTextHelpFormatter = \
        ColorRawTextHelpFormatter         # type: ignore
    ArgumentDefaultsHelpFormatter = \
        ColorArgDefaultsHelpFormatter     # type: ignore
    MetavarTypeHelpFormatter = \
        ColorMetavarTypeHelpFormatter     # type: ignore


def monkey_patch() -> ContextManager:
    """
    Monkey-patch the system argparse module to automatically page help output.

    The result of calling this function can optionally be used as a context
    manager to restore the status quo when it exits.
    """
    import sys

    def get_existing_classes(module: types.ModuleType) -> Tuple[Type, ...]:
        return (
            module._HelpAction,                    # type: ignore
            module.HelpFormatter,                  # type: ignore
            module.RawDescriptionHelpFormatter,    # type: ignore
            module.RawTextHelpFormatter,           # type: ignore
            module.ArgumentDefaultsHelpFormatter,  # type: ignore
            module.MetavarTypeHelpFormatter,       # type: ignore
        )  # type: ignore

    def patch_classes(module: types.ModuleType,
                      impl: Tuple[Type, ...]) -> None:
        (
            module._HelpAction,                    # type: ignore
            module.HelpFormatter,                  # type: ignore
            module.RawDescriptionHelpFormatter,    # type: ignore
            module.RawTextHelpFormatter,           # type: ignore
            module.ArgumentDefaultsHelpFormatter,  # type: ignore
            module.MetavarTypeHelpFormatter,       # type: ignore
        ) = impl

    orig = get_existing_classes(argparse)
    orig_fmtr = argparse.ArgumentParser._get_formatter
    patched = get_existing_classes(sys.modules[__name__])
    patch_classes(argparse, patched)
    new_fmtr = _substitute_formatter(orig_fmtr)
    argparse.ArgumentParser._get_formatter = new_fmtr  # type: ignore

    @contextlib.contextmanager
    def unpatcher() -> Generator:
        try:
            yield
        finally:
            patch_classes(argparse, orig)
            argparse.ArgumentParser._get_formatter = orig_fmtr  # type: ignore

    return unpatcher()


__all__ = argparse.__all__ + [  # type: ignore
        'use_color_for_parser', 'monkey_patch'
    ]



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import contextlib
import fcntl
import itertools
import multiprocessing
import os
import pty
import re
import signal
import struct
import sys
import tempfile
import termios
import time
import traceback

import types
from typing import Optional, Generator, Tuple
import typing


(LINES, COLUMNS) = (24, 80)


def _open_fifo(path: Optional[str],
               write: bool) -> Optional[int]:
    if path is None:
        return None
    return os.open(path, os.O_WRONLY if write else os.O_RDONLY)


class IsolationEnvironment:
    def __init__(self, pid: int, ptm_fd: int,
                 stdin_fifo: Optional[str] = None,
                 stdout_fifo: Optional[str] = None,
                 stderr_fifo: Optional[str] = None):
        self._pid = pid

        self._ptm_fd = ptm_fd
        self._tty = os.fdopen(ptm_fd, 'r')

        self._stdin_fifo_fd = _open_fifo(stdin_fifo, True)
        self._stdout_fifo_fd = _open_fifo(stdout_fifo, False)
        self._stderr_fifo_fd = _open_fifo(stderr_fifo, False)

        self._exit_code: Optional[int] = None

        self._raw_lines: typing.List[str] = []
        self._lines: typing.List[str] = []

    def interrupt(self) -> None:
        self.write(b'\x03')

    def write(self, data: bytes) -> None:
        os.write(self._ptm_fd, data)
        time.sleep(0.001)

    def readline(self) -> str:
        line = self._tty.readline()
        self._raw_lines.append(line)
        return line

    def error_output(self) -> str:
        if self._stderr_fifo_fd is None:
            return ''
        with os.fdopen(self._stderr_fifo_fd) as errf:
            self._stderr_fifo_fd = None
            return errf.read()

    def stdout_pipe(self) -> typing.TextIO:
        assert self._stdout_fifo_fd is not None
        stdout = os.fdopen(self._stdout_fifo_fd)
        self._stdout_fifo_fd = None
        return stdout

    def stdin_pipe(self) -> typing.TextIO:
        assert self._stdin_fifo_fd is not None
        return os.fdopen(self._stdin_fifo_fd, 'w', closefd=False)

    def close(self, get_return_code: typing.Callable[[], int]) -> None:
        for i in range(100):
            if os.waitpid(self._pid, os.WNOHANG) != (0, 0):
                break
            time.sleep(0.001)
        else:
            os.kill(self._pid, signal.SIGTERM)
            os.waitpid(self._pid, 0)
        self._exit_code = get_return_code()
        self._tty.close()
        for fifo_fd in (self._stdin_fifo_fd,
                        self._stdout_fifo_fd,
                        self._stderr_fifo_fd):
            if fifo_fd is not None:
                os.close(fifo_fd)

    def exit_code(self) -> int:
        assert self._exit_code is not None
        return self._exit_code

    def record_output(self, line: str) -> None:
        self._lines.append(line)

    def recorded_output(self) -> typing.Tuple[typing.List[str],
                                              typing.List[str]]:
        return self._raw_lines[:], self._lines[:]


class PagerControl:
    _page_end = None
    _ctrl_chars = re.compile(r'\x1b\['
                             r'(\?|[0-2]?K|[0-9]*;?[0-9]*H|[0-3]?J|[0-9]*m)')

    def __init__(self, isolation_env: IsolationEnvironment):
        self.env = isolation_env
        self._total_lines = 0
        self._lines = self._iter_lines()

    def _iter_lines(self) -> Generator[typing.Union[str, None],
                                       None, None]:
        def get_content(segment: str) -> Tuple[bool,
                                               typing.Union[str, None]]:
            if not segment:
                return False, ''
            if segment == '\x1b[1m~\x1b[0m\n':
                # Ignore lines that are filling blank vertical space with
                # '~' after hitting Ctrl-C when the screen is not full
                return False, ''
            visible = self._ctrl_chars.sub('', segment)
            if ((visible.rstrip() == ':' or '(END)' in visible
                    or 'Waiting for data...' in visible)
                    and segment.replace('\x1b[m', '') != visible):
                return True, self._page_end
            elif visible.rstrip() or segment == visible:
                self._total_lines += 1
                self.env.record_output(visible)
                return True, visible
            return False, ''

        while True:
            line = '\x1b[?'
            while line.lstrip(' q').startswith('\x1b[?'):
                rawline = self.env.readline()
                line = (rawline.replace('\x07', '')     # Ignore bell
                               .replace('\x1b[m', ''))  # Ignore style reset
            before, reset, after = line.partition('\x1b[2J')

            valid, content = get_content(before)
            if valid:
                yield content
            if reset and not (valid and (content is self._page_end)):
                yield self._page_end
            valid, content = get_content(after)
            if valid:
                yield content

    def read_lines(self, count: int) -> typing.Iterator[str]:
        return itertools.islice((line for line in self._lines
                                 if line is not self._page_end), count)

    def _lines_in_page(self) -> int:
        original_count = self._total_lines
        try:
            for line in self._lines:
                if line is self._page_end:
                    break
        except IOError:
            pass
        return self._total_lines - original_count

    def advance(self) -> int:
        self.env.write(b' ')
        return self._lines_in_page()

    def quit(self) -> int:
        self.env.write(b'q')
        return self._lines_in_page()

    def total_lines(self) -> int:
        return self._total_lines


@contextlib.contextmanager
def _fifo(fifo_path: str) -> Generator[str, None, None]:
    try:
        os.mkfifo(fifo_path, 0o600)
        yield fifo_path
    finally:
        try:
            os.unlink(fifo_path)
        except OSError:
            pass


@contextlib.contextmanager
def _fifos(*fifo_names: Optional[str]) -> Generator[typing.List[Optional[str]],
                                                    None, None]:
    with tempfile.TemporaryDirectory() as directory:
        with contextlib.ExitStack() as stack:
            fifos = [stack.enter_context(_fifo(os.path.join(directory,
                                                            name)))
                     if name is not None else None
                     for name in fifo_names]
            yield fifos


class TestProcessNotComplete(Exception):
    pass


class TimeoutException(Exception):
    def __init__(self, pid: Optional[int]):
        super(Exception, self).__init__(f'Test process {pid} timed out')


@contextlib.contextmanager
def isolate(child_function: typing.Callable[[], int],
            stdin_pipe: bool = False,
            stdout_pipe: bool = False,
            stderr_pipe: bool = True,
            *,
            lines: int = LINES,
            columns: int = COLUMNS) -> Generator[IsolationEnvironment,
                                                 None, None]:
    with _fifos('stdin' if stdin_pipe else None,
                'stdout' if stdout_pipe else None,
                'stderr' if stderr_pipe else None) as fifo_paths:
        result_r, result_w = os.pipe()
        env_pid, tty = pty.fork()
        if env_pid == pty.CHILD:
            try:
                os.close(result_r)
                # Get ttyname from original stdout, even if test runner has
                # replaced stdout with another file
                pts = os.ttyname(pty.STDOUT_FILENO)

                ctx = multiprocessing.get_context('spawn')
                p = ctx.Process(target=_run_test, args=(child_function, pts,
                                                        *fifo_paths))
                p.start()

                def handle_terminate(signum: int,
                                     frame: Optional[types.FrameType]) -> None:
                    if p.is_alive():
                        p.terminate()

                signal.signal(signal.SIGTERM, handle_terminate)
                # Signals from the pty are for the test process, not us
                signal.signal(signal.SIGINT, signal.SIG_IGN)

                p.join(2)  # Wait at most 2s
                if p.exitcode is None:
                    raise TimeoutException(p.pid)
            except BaseException:
                try:
                    with os.fdopen(result_w, 'w') as result_writer:
                        traceback.print_exc(file=result_writer)
                finally:
                    # Prevent blocking in parent process by opening our end of
                    # all FIFOs.
                    for path, write in zip(fifo_paths, [False, True, True]):
                        _open_fifo(path, write)
                    os._exit(1)
            with os.fdopen(result_w, 'w') as result_writer:
                result_writer.write(f'{p.exitcode}\n')

            os._exit(0)
        else:
            os.close(result_w)
            fcntl.ioctl(tty, termios.TIOCSWINSZ,
                        struct.pack('HHHH', lines, columns, 0, 0))
            env = IsolationEnvironment(env_pid, tty, *fifo_paths)
            time.sleep(0.01)
            try:
                try:
                    yield env
                finally:
                    def get_return_code() -> int:
                        with os.fdopen(result_r) as result_reader:
                            result = result_reader.readline().rstrip()
                            try:
                                return int(result)
                            except ValueError:
                                pass
                            trace = result_reader.read()
                            raise TestProcessNotComplete('\n'.join([result,
                                                                    trace]))

                    env.close(get_return_code)
            except Exception:
                raw, processed = env.recorded_output()
                if raw:
                    print(f'Raw output: {repr(raw)}', file=sys.stderr)
                if processed:
                    print(f'Recorded output: {repr(processed)}',
                          file=sys.stderr)
                raise


def _run_test(test_function: typing.Callable[[], int],
              pts: str,
              stdin_fifo: Optional[str],
              stdout_fifo: Optional[str],
              stderr_fifo: Optional[str]) -> typing.NoReturn:
    os.environ['TERM'] = 'xterm-256color'

    tty_fd = os.open(pts, os.O_RDWR)
    if stdin_fifo is not None:
        old_stdin, sys.stdin = sys.stdin, open(stdin_fifo)
        old_stdin.close()
    else:
        os.dup2(tty_fd, pty.STDIN_FILENO)
    if stdout_fifo is not None:
        old_stdout, sys.stdout = sys.stdout, open(stdout_fifo, 'w')
        old_stdout.close()
    else:
        os.dup2(tty_fd, pty.STDOUT_FILENO)
    if stderr_fifo is not None:
        old_stderr, sys.stderr = sys.stderr, open(stderr_fifo, 'w',
                                                  buffering=1)
        old_stderr.close()
    else:
        os.dup2(tty_fd, pty.STDERR_FILENO)
    os.close(tty_fd)

    result = test_function()
    sys.exit(result or 0)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import unittest

import fixtures  # type: ignore

from typing import Any, Optional, Dict, List

import autopage
from autopage import command


_PagerConfig = command.PagerConfig


class ConfigTest(unittest.TestCase):
    def setUp(self) -> None:
        class TestCommand(command.PagerCommand):
            def __init__(self) -> None:
                self.config: Optional[_PagerConfig] = None

            def command(self) -> List[str]:
                return []

            def environment_variables(
                    self,
                    config: _PagerConfig) -> Optional[Dict[str, str]]:
                self.config = config
                return None

        self.test_command = TestCommand()

    def _get_ap_config(self, **args: Any) -> command.PagerConfig:
        ap = autopage.AutoPager(pager_command=self.test_command, **args)
        ap._pager_env()
        config = self.test_command.config
        assert config is not None
        return config

    def test_defaults(self) -> None:
        config = self._get_ap_config()
        self.assertTrue(config.color)
        self.assertFalse(config.line_buffering_requested)
        self.assertFalse(config.reset_terminal)

    def test_nocolor(self) -> None:
        config = self._get_ap_config(allow_color=False)
        self.assertFalse(config.color)
        self.assertFalse(config.line_buffering_requested)
        self.assertFalse(config.reset_terminal)

    def test_reset(self) -> None:
        config = self._get_ap_config(reset_on_exit=True)
        self.assertTrue(config.color)
        self.assertFalse(config.line_buffering_requested)
        self.assertTrue(config.reset_terminal)

    def test_linebuffered(self) -> None:
        config = self._get_ap_config(line_buffering=True)
        self.assertTrue(config.color)
        self.assertTrue(config.line_buffering_requested)
        self.assertFalse(config.reset_terminal)

    def test_not_linebuffered(self) -> None:
        config = self._get_ap_config(line_buffering=False)
        self.assertTrue(config.color)
        self.assertFalse(config.line_buffering_requested)
        self.assertFalse(config.reset_terminal)


class EnvironmentBuildTest(unittest.TestCase):
    class TestCommand(command.PagerCommand):
        def __init__(self, env: Optional[Dict[str, str]]):
            self._env = env

        def command(self) -> List[str]:
            return ['foo']

        def environment_variables(self,
                                  config: _PagerConfig) -> Optional[Dict[str,
                                                                         str]]:
            return self._env

    def test_env(self) -> None:
        cmd = self.TestCommand({"FOO": "bar"})
        ap = autopage.AutoPager(pager_command=cmd)
        with fixtures.EnvironmentVariable('BAZ', 'quux'):
            env = ap._pager_env()
        self.assertIsNotNone(env)
        assert env is not None
        self.assertEqual('quux', env['BAZ'])
        self.assertEqual('bar', env['FOO'])

    def test_env_empty(self) -> None:
        cmd = self.TestCommand({})
        ap = autopage.AutoPager(pager_command=cmd)
        with fixtures.EnvironmentVariable('BAZ', 'quux'):
            env = ap._pager_env()
        self.assertIsNone(env)

    def test_env_none(self) -> None:
        cmd = self.TestCommand(None)
        ap = autopage.AutoPager(pager_command=cmd)
        with fixtures.EnvironmentVariable('BAZ', 'quux'):
            env = ap._pager_env()
        self.assertIsNone(env)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import codecs
import subprocess
import sys
import unittest
from unittest import mock

import fixtures  # type: ignore

from typing import Optional, List, Dict

from autopage.tests import sinks

import autopage
from autopage import command


_PagerConfig = command.PagerConfig


class PagedStreamTest(fixtures.TestWithFixtures):
    def setUp(self) -> None:
        out = sinks.TTYFixture()
        self.useFixture(out)
        self.stream = out.stream
        self.useFixture(fixtures.MonkeyPatch('sys.stdout', self.stream))
        popen = fixtures.MockPatch('subprocess.Popen')
        self.useFixture(popen)
        self.popen = popen.mock

    def test_defaults(self) -> None:
        class TestCommand(command.PagerCommand):
            def command(self) -> List[str]:
                return []

            def environment_variables(
                    self,
                    config: _PagerConfig) -> Optional[Dict[str, str]]:
                return None

        tc = TestCommand()
        ap = autopage.AutoPager(pager_command=tc, line_buffering=False)
        with mock.patch.object(ap, '_pager_env') as get_env, \
                mock.patch.object(tc, 'command') as cmd:
            stream = ap._paged_stream()
            self.popen.assert_called_once_with(
                cmd.return_value,
                env=get_env.return_value,
                bufsize=-1,
                universal_newlines=True,
                encoding='utf-8',
                errors='strict',
                stdin=subprocess.PIPE,
                stdout=None)
            self.assertIs(stream, self.popen.return_value.stdin)

    def test_defaults_cmd_as_class(self) -> None:
        class TestCommand(command.PagerCommand):
            def command(self) -> List[str]:
                return []

            def environment_variables(
                    self,
                    config: _PagerConfig) -> Optional[Dict[str, str]]:
                return None

        with mock.patch.object(TestCommand, 'command') as cmd:
            ap = autopage.AutoPager(pager_command=TestCommand,
                                    line_buffering=False)
            with mock.patch.object(ap, '_pager_env') as get_env:
                stream = ap._paged_stream()
                self.popen.assert_called_once_with(
                    cmd.return_value,
                    env=get_env.return_value,
                    bufsize=-1,
                    universal_newlines=True,
                    encoding='utf-8',
                    errors='strict',
                    stdin=subprocess.PIPE,
                    stdout=None)
                self.assertIs(stream, self.popen.return_value.stdin)

    def test_defaults_cmd_as_string(self) -> None:
        ap = autopage.AutoPager(pager_command='foo bar',
                                line_buffering=False)
        with mock.patch.object(ap, '_pager_env') as get_env:
            stream = ap._paged_stream()
            self.popen.assert_called_once_with(
                ['foo', 'bar'],
                env=get_env.return_value,
                bufsize=-1,
                universal_newlines=True,
                encoding='utf-8',
                errors='strict',
                stdin=subprocess.PIPE,
                stdout=None)
            self.assertIs(stream, self.popen.return_value.stdin)

    def test_defaults_cmd_as_int(self) -> None:
        self.assertRaises(TypeError, autopage.AutoPager,
                          pager_command=42)

    def test_line_buffering(self) -> None:
        ap = autopage.AutoPager(line_buffering=True)
        stream = ap._paged_stream()
        self.popen.assert_called_once_with(
            mock.ANY,
            env=mock.ANY,
            bufsize=1,
            universal_newlines=True,
            encoding=mock.ANY,
            errors=mock.ANY,
            stdin=subprocess.PIPE,
            stdout=None)
        self.assertIs(stream, self.popen.return_value.stdin)

    def test_errors(self) -> None:
        ap = autopage.AutoPager(errors=autopage.ErrorStrategy.NAME_REPLACE)
        stream = ap._paged_stream()
        self.popen.assert_called_once_with(
            mock.ANY,
            env=mock.ANY,
            bufsize=mock.ANY,
            universal_newlines=mock.ANY,
            encoding=mock.ANY,
            errors='namereplace',
            stdin=subprocess.PIPE,
            stdout=None)
        self.assertIs(stream, self.popen.return_value.stdin)

    def test_explicit_stdout_stream(self) -> None:
        ap = autopage.AutoPager(self.stream)
        stream = ap._paged_stream()
        self.popen.assert_called_once_with(
            mock.ANY,
            env=mock.ANY,
            bufsize=mock.ANY,
            universal_newlines=mock.ANY,
            encoding=mock.ANY,
            errors=mock.ANY,
            stdin=subprocess.PIPE,
            stdout=None)
        self.assertIs(stream, self.popen.return_value.stdin)

    def test_explicit_stream(self) -> None:
        with sinks.TTYFixture() as tty:
            ap = autopage.AutoPager(tty.stream)
            stream = ap._paged_stream()
            self.popen.assert_called_once_with(
                mock.ANY,
                env=mock.ANY,
                bufsize=mock.ANY,
                universal_newlines=mock.ANY,
                encoding=mock.ANY,
                errors=mock.ANY,
                stdin=subprocess.PIPE,
                stdout=tty.stream)
            self.assertIs(stream, self.popen.return_value.stdin)


class ToTerminalTest(unittest.TestCase):
    def test_pty(self) -> None:
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            self.assertTrue(ap.to_terminal())

    def test_stringio(self) -> None:
        with sinks.BufferFixture() as out:
            ap = autopage.AutoPager(out.stream)
            self.assertFalse(ap.to_terminal())

    def test_file(self) -> None:
        with sinks.TempFixture() as out:
            ap = autopage.AutoPager(out.stream)
            self.assertFalse(ap.to_terminal())

    def test_default_pty(self) -> None:
        with sinks.TTYFixture() as out:
            with fixtures.MonkeyPatch('sys.stdout', out.stream):
                ap = autopage.AutoPager()
            self.assertTrue(ap.to_terminal())

    def test_default_file(self) -> None:
        with sinks.TempFixture() as out:
            with fixtures.MonkeyPatch('sys.stdout', out.stream):
                ap = autopage.AutoPager()
            self.assertFalse(ap.to_terminal())

    def test_launch_pager(self) -> None:
        ap = autopage.AutoPager()
        with mock.patch.object(ap, 'to_terminal', return_value=True), \
                mock.patch.object(ap, '_paged_stream') as page, \
                mock.patch.object(ap, '_reconfigure_output_stream') as reconf:
            with ap as stream:
                page.assert_called_once()
                self.assertIs(page.return_value, stream)
                reconf.assert_not_called()

    def test_launch_pager_fail(self) -> None:
        outstream = mock.Mock()
        ap = autopage.AutoPager(outstream)
        with mock.patch.object(ap, 'to_terminal', return_value=True), \
                mock.patch.object(ap, '_paged_stream',
                                  side_effect=OSError) as page, \
                mock.patch.object(ap, '_reconfigure_output_stream') as reconf:
            with ap as stream:
                page.assert_called_once()
                reconf.assert_called_once()
                self.assertIs(outstream, stream)

    def test_no_pager(self) -> None:
        outstream = mock.Mock()
        ap = autopage.AutoPager(outstream)
        with mock.patch.object(ap, 'to_terminal', return_value=False), \
                mock.patch.object(ap, '_paged_stream') as page, \
                mock.patch.object(ap, '_reconfigure_output_stream') as reconf:
            with ap as stream:
                page.assert_not_called()
                self.assertIs(outstream, stream)
                reconf.assert_called_once()

    def test_pager_cat(self) -> None:
        outstream = mock.Mock()
        cat = command.CustomPager('cat')
        ap = autopage.AutoPager(outstream, pager_command=cat)
        with mock.patch.object(ap, 'to_terminal', return_value=True), \
                mock.patch.object(ap, '_paged_stream') as page, \
                mock.patch.object(ap, '_reconfigure_output_stream') as reconf:
            with ap as stream:
                page.assert_not_called()
                self.assertIs(outstream, stream)
                reconf.assert_called_once()


class ExitCodeTest(fixtures.TestWithFixtures):
    def setUp(self) -> None:
        out = sinks.BufferFixture()
        self.useFixture(out)
        self.ap = autopage.AutoPager(out.stream)

    def test_success(self) -> None:
        with self.ap:
            pass
        self.assertEqual(0, self.ap.exit_code())

    def test_pager_broken_pipe_flush(self) -> None:
        flush = mock.MagicMock(side_effect=BrokenPipeError)
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            with fixtures.MockPatch('subprocess.Popen') as popen:
                with sinks.BufferFixture() as pager_in:
                    popen.mock.return_value.stdin = pager_in.stream
                    with ap as stream:
                        stream.write('foo')
                        stream.close = flush  # type: ignore
            self.assertEqual(141, ap.exit_code())

    def test_no_pager_broken_pipe_flush(self) -> None:
        flush = mock.MagicMock(side_effect=BrokenPipeError)
        with self.ap as stream:
            stream.write('foo')
            stream.flush = flush  # type: ignore
        self.assertEqual(141, self.ap.exit_code())

    def test_broken_pipe(self) -> None:
        with self.ap:
            raise BrokenPipeError
        self.assertEqual(141, self.ap.exit_code())

    def test_exception(self) -> None:
        class MyException(Exception):
            pass

        def run() -> None:
            with self.ap:
                raise MyException

        self.assertRaises(MyException, run)
        self.assertEqual(1, self.ap.exit_code())

    def test_exception_on_error_default(self) -> None:
        class MyException(Exception):
            pass

        def run() -> None:
            with self.ap:
                raise MyException

        self.assertRaises(MyException, run)
        self.assertEqual(2, self.ap.exit_code(on_error_default=2))

    def test_base_exception(self) -> None:
        class MyBaseException(BaseException):
            pass

        def run() -> None:
            with self.ap:
                raise MyBaseException

        self.assertRaises(MyBaseException, run)
        self.assertEqual(1, self.ap.exit_code())

    def test_interrupt(self) -> None:
        def run() -> None:
            with self.ap:
                raise KeyboardInterrupt

        self.assertRaises(KeyboardInterrupt, run)
        self.assertEqual(130, self.ap.exit_code())

    def test_system_exit(self) -> None:
        def run() -> None:
            with self.ap:
                raise SystemExit(42)

        self.assertRaises(SystemExit, run)
        self.assertEqual(42, self.ap.exit_code())

    def test_system_exit_on_error_default(self) -> None:
        def run() -> None:
            with self.ap:
                raise SystemExit(42)

        self.assertRaises(SystemExit, run)
        self.assertEqual(42, self.ap.exit_code(on_error_default=3))

    def test_system_exit_string(self) -> None:
        def run() -> None:
            with self.ap:
                raise SystemExit("nope")

        self.assertRaises(SystemExit, run)
        self.assertEqual(1, self.ap.exit_code())

    def test_system_exit_string_on_error_default(self) -> None:
        def run() -> None:
            with self.ap:
                raise SystemExit("nope")

        self.assertRaises(SystemExit, run)
        self.assertEqual(3, self.ap.exit_code(on_error_default=3))


class CleanupTest(unittest.TestCase):
    def test_no_pager_stream_not_closed(self) -> None:
        flush = mock.MagicMock()
        with sinks.BufferFixture() as out:
            with autopage.AutoPager(out.stream) as stream:
                stream.flush = flush  # type: ignore
                stream.write('foo')
            self.assertFalse(out.stream.closed)
        flush.assert_called_once()

    def test_no_pager_broken_pipe(self) -> None:
        flush = mock.MagicMock(side_effect=BrokenPipeError)
        with sinks.BufferFixture() as out:
            with autopage.AutoPager(out.stream) as stream:
                stream.flush = flush  # type: ignore
                stream.write('foo')
            self.assertTrue(out.stream.closed)
        flush.assert_called_once()

    def test_no_pager_broken_pipe_flush(self) -> None:
        flush = mock.MagicMock(side_effect=BrokenPipeError)
        with sinks.BufferFixture() as out:
            with autopage.AutoPager(out.stream) as stream:
                stream.write('foo')
                stream.flush = flush  # type: ignore
            self.assertTrue(out.stream.closed)
        flush.assert_called_once()

    def test_no_pager_stream_closed(self) -> None:
        flush = mock.MagicMock(side_effect=ValueError)
        with sinks.BufferFixture() as out:
            with autopage.AutoPager(out.stream) as stream:
                stream.write('foo')
                stream.close()
                # Calling flush() on a closed stream raises an exception for
                # real streams (but not for StringIO).
                stream.flush = flush  # type: ignore
            self.assertTrue(out.stream.closed)

    def test_pager_stream_not_closed(self) -> None:
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            with fixtures.MockPatch('subprocess.Popen') as popen:
                with sinks.BufferFixture() as pager_in:
                    popen.mock.return_value.stdin = pager_in.stream
                    with ap as stream:
                        self.assertIs(pager_in.stream, stream)
                    self.assertTrue(pager_in.stream.closed)

    def test_pager_stream_not_closed_interrupt(self) -> None:
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            with fixtures.MockPatch('subprocess.Popen') as popen:
                with sinks.BufferFixture() as pager_in:
                    popen.mock.return_value.stdin = pager_in.stream

                    def run() -> None:
                        with ap as stream:
                            self.assertIs(pager_in.stream, stream)
                            raise KeyboardInterrupt

                    self.assertRaises(KeyboardInterrupt, run)
                    self.assertTrue(pager_in.stream.closed)

    def test_pager_broken_pipe(self) -> None:
        flush = mock.MagicMock(side_effect=BrokenPipeError)
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            with fixtures.MockPatch('subprocess.Popen') as popen:
                with sinks.BufferFixture() as pager_in:
                    popen.mock.return_value.stdin = pager_in.stream
                    pager_in.stream.flush = flush
                    with ap as stream:
                        self.assertIs(pager_in.stream, stream)
                    self.assertTrue(pager_in.stream.closed)
                    popen.mock.return_value.wait.assert_called_once()

    def test_pager_stream_closed(self) -> None:
        with sinks.TTYFixture() as out:
            ap = autopage.AutoPager(out.stream)
            with fixtures.MockPatch('subprocess.Popen') as popen:
                with sinks.BufferFixture() as pager_in:
                    popen.mock.return_value.stdin = pager_in.stream
                    with ap as stream:
                        self.assertIs(pager_in.stream, stream)
                        stream.close()
                    popen.mock.return_value.wait.assert_called_once()


class StreamConfigureTest(fixtures.TestWithFixtures):
    def setUp(self) -> None:
        out = sinks.TempFixture()
        self.useFixture(out)
        self.stream = out.stream
        self.default_lb = self.stream.line_buffering
        self.default_errors = self.stream.errors
        self.encoding = codecs.lookup(self.stream.encoding).name

    def test_line_buffering_on(self) -> None:
        ap = autopage.AutoPager(self.stream, line_buffering=True)
        ap._reconfigure_output_stream()
        self.addCleanup(ap._out.close)
        self.assertTrue(ap._out.line_buffering)
        self.assertEqual(self.default_errors, ap._out.errors)
        self.assertEqual(self.encoding, ap._encoding())
        self.assertIs(True, ap._line_buffering())
        self.assertEqual(self.default_errors, ap._errors())

    def test_line_buffering_off(self) -> None:
        ap = autopage.AutoPager(self.stream, line_buffering=False)
        ap._reconfigure_output_stream()
        self.addCleanup(ap._out.close)
        self.assertFalse(ap._out.line_buffering)
        self.assertEqual(self.default_errors, ap._out.errors)
        self.assertEqual(self.encoding, ap._encoding())
        self.assertIs(False, ap._line_buffering())
        self.assertEqual(self.default_errors, ap._errors())

    def test_stdout_line_buffering_on(self) -> None:
        with fixtures.MonkeyPatch('sys.stdout', self.stream):
            ap = autopage.AutoPager(line_buffering=True)
            ap._reconfigure_output_stream()
            self.addCleanup(ap._out.close)
            self.assertTrue(sys.stdout.line_buffering)
            self.assertEqual(self.default_errors, sys.stdout.errors)
            self.assertEqual(self.encoding,
                             codecs.lookup(sys.stdout.encoding).name)

    def test_errors(self) -> None:
        ap = autopage.AutoPager(self.stream,
                                errors=autopage.ErrorStrategy.NAME_REPLACE)
        ap._reconfigure_output_stream()
        self.addCleanup(ap._out.close)
        self.assertEqual(self.default_lb, ap._out.line_buffering)
        self.assertEqual('namereplace', ap._out.errors)
        self.assertNotEqual(self.default_errors, ap._out.errors)
        self.assertEqual(self.encoding, ap._encoding())
        self.assertEqual('namereplace', ap._errors())
        self.assertEqual(self.default_lb, ap._line_buffering())

    def test_errors_string(self) -> None:
        ap = autopage.AutoPager(self.stream,
                                errors='namereplace')  # type: ignore
        ap._reconfigure_output_stream()
        self.addCleanup(ap._out.close)
        self.assertEqual(self.default_lb, ap._out.line_buffering)
        self.assertEqual('namereplace', ap._out.errors)
        self.assertNotEqual(self.default_errors, ap._out.errors)
        self.assertEqual(self.encoding, ap._encoding())
        self.assertEqual('namereplace', ap._errors())
        self.assertEqual(self.default_lb, ap._line_buffering())

    def test_errors_bogus_string(self) -> None:
        self.assertRaises(ValueError,
                          autopage.AutoPager,
                          self.stream, errors='panic')

    def test_line_buffering_on_errors(self) -> None:
        ap = autopage.AutoPager(self.stream,
                                line_buffering=True,
                                errors=autopage.ErrorStrategy.NAME_REPLACE)
        ap._reconfigure_output_stream()
        self.addCleanup(ap._out.close)
        self.assertTrue(ap._out.line_buffering)
        self.assertEqual('namereplace', ap._out.errors)
        self.assertNotEqual(self.default_errors, ap._out.errors)
        self.assertEqual(self.encoding, ap._encoding())
        self.assertIs(True, ap._line_buffering())
        self.assertEqual('namereplace', ap._errors())



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import unittest
import sys

import fixtures  # type: ignore

import typing

from autopage import command


class MoreTest(unittest.TestCase):
    def setUp(self) -> None:
        self.cmd = command.More()

    def test_cmd(self) -> None:
        self.assertEqual(['more'], self.cmd.command())

    def test_less_env_defaults(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        self.assertIsNone(self.cmd.environment_variables(config))


class LessTest(unittest.TestCase):
    def setUp(self) -> None:
        self.cmd: command.PagerCommand = command.Less()

    def _env(self, config: command.PagerConfig) -> typing.Dict:
        return self.cmd.environment_variables(config) or {}

    def test_cmd(self) -> None:
        self.assertEqual(['less'], self.cmd.command())

    def test_less_env_defaults(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        less_env = self._env(config)['LESS']
        self.assertEqual('RFX', less_env)

    def test_less_env_nocolor(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        less_env = self._env(config)['LESS']
        self.assertEqual('FX', less_env)

    def test_less_env_reset(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=True)
        less_env = self._env(config)['LESS']
        self.assertEqual('R', less_env)

    def test_less_env_nocolor_reset(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=False,
                                     reset_terminal=True)
        self.assertNotIn('LESS', self._env(config))

    def test_less_env_linebuffered(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=True,
                                     reset_terminal=False)
        less_env = self._env(config)['LESS']
        self.assertEqual('RX', less_env)

    def test_less_env_linebuffered_reset(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=True,
                                     reset_terminal=True)
        less_env = self._env(config)['LESS']
        self.assertEqual('R', less_env)

    def test_less_env_linebuffered_nocolor(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=True,
                                     reset_terminal=False)
        less_env = self._env(config)['LESS']
        self.assertEqual('X', less_env)

    def test_less_env_linebuffered_nocolor_reset(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=True,
                                     reset_terminal=True)
        self.assertNotIn('LESS', self._env(config))

    def test_less_env_override(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        with fixtures.EnvironmentVariable('LESS', 'abc'):
            env = self._env(config)
        self.assertNotIn('LESS', env)


class LVTest(unittest.TestCase):
    def setUp(self) -> None:
        self.cmd: command.PagerCommand = command.LV()

    def _env(self, config: command.PagerConfig) -> typing.Dict:
        return self.cmd.environment_variables(config) or {}

    def test_cmd(self) -> None:
        self.assertEqual(['lv'], self.cmd.command())

    def test_lv_env_defaults(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        lv_env = self._env(config)['LV']
        self.assertEqual('-c', lv_env)

    def test_lv_env_nocolor(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        self.assertNotIn('LV', self._env(config))

    def test_lv_env_reset(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=True)
        lv_env = self._env(config)['LV']
        self.assertEqual('-c', lv_env)

    def test_lv_env_nocolor_reset(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=False,
                                     reset_terminal=True)
        self.assertNotIn('LV', self._env(config))

    def test_lv_env_linebuffered(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=True,
                                     reset_terminal=False)
        lv_env = self._env(config)['LV']
        self.assertEqual('-c', lv_env)

    def test_lv_env_linebuffered_reset(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=True,
                                     reset_terminal=True)
        lv_env = self._env(config)['LV']
        self.assertEqual('-c', lv_env)

    def test_lv_env_linebuffered_nocolor(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=True,
                                     reset_terminal=False)
        self.assertNotIn('LV', self._env(config))

    def test_lv_env_linebuffered_nocolor_reset(self) -> None:
        config = command.PagerConfig(color=False,
                                     line_buffering_requested=True,
                                     reset_terminal=True)
        self.assertNotIn('LV', self._env(config))

    def test_lv_env_override(self) -> None:
        config = command.PagerConfig(color=True,
                                     line_buffering_requested=False,
                                     reset_terminal=False)
        with fixtures.EnvironmentVariable('LV', 'abc'):
            env = self._env(config)
        self.assertNotIn('LV', env)


class DefaultTest(LessTest, LVTest):
    def setUp(self) -> None:
        with fixtures.EnvironmentVariable('PAGER', 'less "-r" +F'):
            self.cmd = command.DefaultPager()

    def test_cmd(self) -> None:
        self.assertEqual(['less', '-r', '+F'], self.cmd.command())

    def test_default_cmd(self) -> None:
        with fixtures.EnvironmentVariable('PAGER'):
            cmd = command.DefaultPager()
        self.assertEqual(command.PlatformPager().command(),
                         cmd.command())


class UserSpecifiedTest(DefaultTest):
    def setUp(self) -> None:
        with fixtures.EnvironmentVariable('FOO_PAGER', 'less "-r" +F'):
            self.cmd = command.UserSpecifiedPager('FOO_PAGER')

    def test_env_var_priority_cmd(self) -> None:
        with fixtures.EnvironmentVariable('FOO', 'foo'):
            cmd = command.UserSpecifiedPager('FOO', 'BAR')
        self.assertEqual(['foo'], cmd.command())

    def test_env_var_fallthrough_cmd(self) -> None:
        with fixtures.EnvironmentVariable('BAR', 'bar'):
            cmd = command.UserSpecifiedPager('FOO', 'BAR')
        self.assertEqual(['bar'], cmd.command())

    def test_default_cmd(self) -> None:
        with fixtures.EnvironmentVariable('FOO'):
            with fixtures.EnvironmentVariable('BAR'):
                cmd = command.UserSpecifiedPager('FOO', 'BAR')
        self.assertEqual(command.PlatformPager().command(),
                         cmd.command())


class PlatformFixture(fixtures.Fixture):
    def __init__(self, platform: str):
        self.platform = platform

    def _setUp(self) -> None:
        self.addCleanup(setattr, sys, 'platform', sys.platform)
        sys.platform = self.platform


class PlatformTest(unittest.TestCase):
    def test_aix_cmd(self) -> None:
        with PlatformFixture('aix'):
            cmd = command.PlatformPager()
            self.assertEqual(['more'], cmd.command())

        # Prior to Python 3.8, the version number was included
        with PlatformFixture('aix7'):
            cmd = command.PlatformPager()
            self.assertEqual(['more'], cmd.command())

    def test_linux_cmd(self) -> None:
        with PlatformFixture('linux'):
            cmd = command.PlatformPager()
            self.assertEqual(['less'], cmd.command())

    def test_win32_cmd(self) -> None:
        with PlatformFixture('win32'):
            cmd = command.PlatformPager()
            self.assertEqual(['more.com'], cmd.command())

    def test_cygwin_cmd(self) -> None:
        with PlatformFixture('cygwin'):
            cmd = command.PlatformPager()
            self.assertEqual(['less'], cmd.command())

    def test_macos_cmd(self) -> None:
        with PlatformFixture('darwin'):
            cmd = command.PlatformPager()
            self.assertEqual(['less'], cmd.command())

    def test_sunos_cmd(self) -> None:
        with PlatformFixture('sunos5'):
            cmd = command.PlatformPager()
            self.assertEqual(['less'], cmd.command())

    def test_freebsd_cmd(self) -> None:
        with PlatformFixture('freebsd8'):
            cmd = command.PlatformPager()
            self.assertEqual(['less'], cmd.command())


class GetPagerCommandTest(unittest.TestCase):
    def test_instance(self) -> None:
        cmd = command.PlatformPager()
        self.assertIs(cmd, command.get_pager_command(cmd))

    def test_subclass(self) -> None:
        cls = command.Less
        self.assertIsInstance(command.get_pager_command(cls), cls)

    def test_func(self) -> None:
        func = command.PlatformPager
        self.assertIsInstance(command.get_pager_command(func), type(func()))

    def test_string(self) -> None:
        cmd = command.get_pager_command('foo bar')
        self.assertIsInstance(cmd, command.CustomPager)
        self.assertEqual(['foo', 'bar'], cmd.command())

    def test_list(self) -> None:
        with fixtures.EnvironmentVariable('FOO', 'foo'):
            cmd = command.get_pager_command(['FOO', 'BAR'])
        self.assertIsInstance(cmd, command.CustomPager)
        self.assertEqual(['foo'], cmd.command())

    def test_int(self) -> None:
        self.assertRaises(TypeError, command.get_pager_command, 42)

    def test_list_int(self) -> None:
        self.assertRaises(TypeError, command.get_pager_command, ['FOO', 42])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import io
import os
import pty
import tempfile

import fixtures  # type: ignore

import typing


class TTYFixture(fixtures.Fixture):
    def _setUp(self) -> None:
        self._k, term = pty.openpty()
        self.stream = os.fdopen(term, 'w')
        self.addCleanup(self.stream.close)
        self.addCleanup(os.close, self._k)


class TempFixture(fixtures.Fixture):
    def _setUp(self) -> None:
        self.stream = typing.cast(typing.TextIO, tempfile.TemporaryFile('w'))

        def close() -> None:
            try:
                self.stream.close()
            except ValueError:
                pass

        self.addCleanup(close)


class BufferFixture(fixtures.Fixture):
    def _setUp(self) -> None:
        self.stream: typing.TextIO = io.StringIO()
        self.addCleanup(self.stream.close)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
A library to provide automatic paging for console output.

By Zane Bitter.
"""

import codecs
import contextlib
import enum
import io
import os
import signal
import subprocess
import sys

import types
import typing
from typing import Any, Optional, Type, Dict, TextIO

from autopage import command


__all__ = ['AutoPager', 'line_buffer_from_input']


class ErrorStrategy(enum.Enum):
    """
    The strategy for dealing with unicode errors when convering text to bytes.
    """
    STRICT = 'strict'
    IGNORE = 'ignore'
    REPLACE = 'replace'
    BACKSLASH_REPLACE = 'backslashreplace'
    XML_CHARREF_REPLACE = 'xmlcharrefreplace'
    NAME_REPLACE = 'namereplace'


class AutoPager:
    """
    A context manager that launches a pager for the output if appropriate.

    If the output stream is not to the console (i.e. it is piped or
    redirected), no pager will be launched.
    """

    def __init__(self,
                 output_stream: Optional[TextIO] = None, *,
                 pager_command: command.CommandType = command.DefaultPager(),
                 allow_color: bool = True,
                 line_buffering: Optional[bool] = None,
                 reset_on_exit: bool = False,
                 errors: Optional[ErrorStrategy] = None):
        self._use_stdout = output_stream is None or output_stream is sys.stdout
        self._out = sys.stdout if output_stream is None else output_stream
        self._tty = (not self._out.closed) and self._out.isatty()
        self._command = command.get_pager_command(pager_command)
        self._config = command.PagerConfig(
                color=allow_color,
                line_buffering_requested=bool(line_buffering),
                reset_terminal=reset_on_exit,
            )
        self._set_line_buffering = line_buffering
        self._set_errors = (ErrorStrategy(errors) if errors is not None
                            else None)
        self._pager: Optional[subprocess.Popen] = None
        self._exit_code: typing.Union[int, BaseException] = 0

    def to_terminal(self) -> bool:
        """Return whether the output stream is a terminal."""
        return self._tty

    def __enter__(self) -> TextIO:
        # Only invoke the pager if the output is going to a tty; if it is
        # being sent to a file or pipe then we don't want the pager involved
        if self.to_terminal() and self._command.command() != ['cat']:
            try:
                return self._paged_stream()
            except OSError:
                pass
        self._reconfigure_output_stream()
        return self._out

    def _line_buffering(self) -> bool:
        if self._set_line_buffering is None:
            return getattr(self._out, 'line_buffering', self._tty)
        return self._set_line_buffering

    def _encoding(self) -> str:
        dflt = 'utf-8' if getattr(sys.flags, 'utf8_mode', False) else 'ascii'
        encoding = getattr(self._out, 'encoding', dflt)
        try:
            return codecs.lookup(encoding).name
        except LookupError:
            return dflt

    def _errors(self) -> str:
        if self._set_errors is None:
            return getattr(self._out, 'errors', ErrorStrategy.STRICT.value)
        return self._set_errors.value

    def _reconfigure_output_stream(self) -> None:
        if self._set_line_buffering is None and self._set_errors is None:
            return

        if not isinstance(self._out, io.TextIOWrapper):
            return

        # Python 3.7 & later
        if hasattr(self._out, 'reconfigure'):
            self._out.reconfigure(line_buffering=self._set_line_buffering,
                                  errors=(self._set_errors.value
                                          if self._set_errors is not None
                                          else None))
        # Python 3.6
        elif (self._out.line_buffering != self._line_buffering()
                or self._out.errors != self._errors()):
            # Pure-python I/O
            if (hasattr(self._out, '_line_buffering')
                    and hasattr(self._out, '_errors')):
                py_out = typing.cast(Any, self._out)
                py_out._line_buffering = self._line_buffering()
                py_out._errors = self._errors()
                py_out.flush()
            # Native C I/O
            else:
                encoding = self._encoding()
                errors = self._errors()
                line_buffering = self._line_buffering()
                try:
                    if self._use_stdout:
                        sys.stdout = typing.cast(TextIO, None)
                    newstream = io.TextIOWrapper(
                        self._out.detach(),
                        line_buffering=line_buffering,
                        encoding=encoding,
                        errors=errors)
                    self._out = newstream
                finally:
                    if self._use_stdout:
                        sys.stdout = self._out

    def _pager_env(self) -> Optional[Dict[str, str]]:
        new_vars = self._command.environment_variables(self._config)
        if not new_vars:
            return None

        env = dict(os.environ)
        env.update(new_vars)
        return env

    def _pager_out_stream(self) -> Optional[TextIO]:
        if not self._use_stdout:
            try:
                # Ensure the output stream has a file descriptor
                self._out.fileno()
            except OSError:
                pass
            else:
                return self._out
        return None

    def _paged_stream(self) -> TextIO:
        buffer_size = 1 if self._line_buffering() else -1
        self._pager = subprocess.Popen(self._command.command(),
                                       env=self._pager_env(),
                                       bufsize=buffer_size,
                                       universal_newlines=True,
                                       encoding=self._encoding(),
                                       errors=self._errors(),
                                       stdin=subprocess.PIPE,
                                       stdout=self._pager_out_stream())
        assert self._pager.stdin is not None
        return typing.cast(TextIO, self._pager.stdin)

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc: Optional[BaseException],
                 traceback: Optional[types.TracebackType]) -> bool:
        if self._pager is not None:
            # Pager ignores Ctrl-C, so we should too
            with _sigint_ignore():
                pager_in = self._pager.stdin
                assert pager_in is not None
                try:
                    pager_in.close()
                except BrokenPipeError:
                    # Other end of pipe already closed
                    self._exit_code = _signal_exit_code(signal.SIGPIPE)
                # Wait for user to exit pager
                self._pager.wait()
        else:
            self._flush_output()
        return self._process_exception(exc)

    def _flush_output(self) -> None:
        try:
            if not self._out.closed:
                self._out.flush()
        except BrokenPipeError:
            self._exit_code = _signal_exit_code(signal.SIGPIPE)
            try:
                # Other end of pipe already closed, so close the stream now
                # and handle the error. If we leave the stream open with
                # unflushed data, then it will print an unhandleable
                # exception during Python's interpreter shutdown.
                self._out.close()
            except BrokenPipeError:
                # This will always happen
                pass

    def _process_exception(self, exc: Optional[BaseException]) -> bool:
        if exc is not None:
            if isinstance(exc, BrokenPipeError):
                self._exit_code = _signal_exit_code(signal.SIGPIPE)
                # Suppress exceptions caused by a broken pipe (indicating that
                # the user has exited the pager, or the following process in
                # the pipeline has exited)
                return True
            elif isinstance(exc, KeyboardInterrupt):
                self._exit_code = _signal_exit_code(signal.SIGINT)
            elif isinstance(exc, SystemExit) and isinstance(exc.code, int):
                self._exit_code = exc.code
            else:
                self._exit_code = exc
        return False

    def exit_code(self, *, on_error_default: int = 1) -> int:
        """
        Return an appropriate exit code for the process based on any errors.

        If the user exits the program prematurely by closing the pager, we may
        want to return a different exit code for the process. This method
        returns an appropriate exit code on the basis of the existence and type
        of any uncaught exceptions.
        """
        if not isinstance(self._exit_code, int):
            return on_error_default
        return self._exit_code


def line_buffer_from_input(input_stream: Optional[typing.IO] = None) -> bool:
    """
    Return whether line buffering should be enabled for a given input stream.

    When data is being read from an input stream, processed somehow, and then
    written to an autopaged output stream, it may be desirable to enable line
    buffering on the output. This means that each line of data written to the
    output will be visible immediately, as opposed to waiting for the output
    buffer to fill up. This is, however, slower.

    If the input stream is a file, line buffering is unnecessary. This function
    determines whether an input stream might require line buffering on output.

    If no input stream is specified, sys.stdin is assumed.

        >>> with AutoPager(line_buffering=line_buffer_from_input()) as out:
        >>>     for l in sys.stdin:
        >>>         out.write(l)
    """
    if input_stream is None:
        input_stream = sys.stdin
    # On Illumos, TTYs claim to be seekable so don't believe them
    return not (input_stream.seekable() and not input_stream.isatty())


def _signal_exit_code(signum: signal.Signals) -> int:
    """
    Return the exit code corresponding to a received signal.

    Conventionally, when a program exits due to a signal its exit code is 128
    plus the signal number.
    """
    return 128 + int(signum)


@contextlib.contextmanager
def _sigint_ignore() -> typing.Generator[None, None, None]:
    """
    Context manager to temporarily ignore SIGINT.
    """
    old_int_handler: Any = None
    try:
        old_int_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        yield
    finally:
        # If this is called from a finalizer during interpreter shutdown,
        # CPython will have removed the definition of SIG_IGN, so we can't
        # set the signal handler back to anything. We can detect this by
        # checking for None returned from getsignal()
        if signal.getsignal(signal.SIGINT) is not None:
            signal.signal(signal.SIGINT, old_int_handler)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import abc
import collections
import collections.abc
import os
import sys

import typing
from typing import Optional, Dict, List


__all__ = ['DefaultPager', 'UserSpecifiedPager', 'PlatformPager',
           'CustomPager',
           'More', 'Less', 'LV']


PagerConfig = collections.namedtuple('PagerConfig', [
        'color',
        'line_buffering_requested',
        'reset_terminal',
    ])


class PagerCommand(metaclass=abc.ABCMeta):
    """
    Abstract base class for pager commands.

    A subclass implementing this interface can be used to specify a particular
    pager command to run and its environment.
    """

    @abc.abstractmethod
    def command(self) -> List[str]:
        """Return the list of command arguments."""
        return ['more']

    @abc.abstractmethod
    def environment_variables(self,
                              config: PagerConfig) -> Optional[Dict[str, str]]:
        """Return the dict of any environment variables to set."""
        return None


class More(PagerCommand):
    """The pager command ``more``."""

    def command(self) -> List[str]:
        if sys.platform.startswith('win32'):
            return ['more.com']
        return ['more']

    def environment_variables(self,
                              config: PagerConfig) -> Optional[Dict[str, str]]:
        return None


class Less(PagerCommand):
    """The pager command ``less``."""

    def command(self) -> List[str]:
        return ['less']

    def environment_variables(self,
                              config: PagerConfig) -> Optional[Dict[str, str]]:
        flags = []
        if config.color:
            # This option will cause less to output ANSI color escape sequences
            # in raw form.
            # Equivalent to the --RAW-CONTROL-CHARS argument
            flags.append('R')
        if not config.line_buffering_requested and not config.reset_terminal:
            # This option will cause less to buffer until an entire screen's
            # worth of data is available (or the EOF is reached), so don't
            # enable it when line buffering is explicitly requested. It also
            # does not reset the terminal after exiting, so don't enable it
            # when resetting the terminal is requested.
            # Equivalent to the --quit-if-one-screen argument
            flags.append('F')
        if not config.reset_terminal:
            # This option will cause less to not reset the terminal after
            # exiting.
            # Equivalent to the --no-init argument
            flags.append('X')

        if flags and (os.getenv('LESS') is None):
            return {
                'LESS': ''.join(flags)
            }
        return None


class LV(PagerCommand):
    """The pager command ``lv``."""

    def command(self) -> List[str]:
        return ['lv']

    def environment_variables(self,
                              config: PagerConfig) -> Optional[Dict[str, str]]:
        if config.color and (os.getenv('LV') is None):
            return {
                # This option allows ANSI color escape sequences in lv
                'LV': '-c',
            }
        return None


class CustomPager(PagerCommand):
    """A pager command parsed from a user-specified string."""
    def __init__(self, pager_cmdline: str):
        import shlex
        self._cmd = shlex.split(pager_cmdline)

    def command(self) -> List[str]:
        return self._cmd

    def environment_variables(self,
                              config: PagerConfig) -> Optional[Dict[str, str]]:
        env = {}
        for cmd_provider in (Less, LV):
            cmd_env = cmd_provider().environment_variables(config)
            if cmd_env:
                env.update(cmd_env)
        return env or None


def PlatformPager() -> PagerCommand:
    """
    Return the default pager command for the current platform.
    """
    if sys.platform.startswith('aix'):
        return More()
    if sys.platform.startswith('win32'):
        return More()
    return Less()


def UserSpecifiedPager(*env_vars: str) -> PagerCommand:
    """
    Return the pager command for the current environment.

    Each of the specified environment variables is searched in order; the first
    one that is set will be used as the pager command. If none of the
    environment variables is set, the default pager for the platform will be
    used.
    """
    for env_var in env_vars:
        env_pager = os.getenv(env_var)
        if env_pager:
            return CustomPager(env_pager)
    return PlatformPager()


def DefaultPager() -> PagerCommand:
    """
    Return the default pager command for the current environment.

    If there is a $PAGER environment variable configured, this command will be
    used. Otherwise, the default pager for the platform will be used.
    """

    if os.name == 'posix':
        # Check that we are not running with elevated privileges
        if os.getuid() not in {0, os.geteuid()}:
            return PlatformPager()
    return UserSpecifiedPager('PAGER')


CommandType = typing.Union[PagerCommand,
                           typing.Callable[[], PagerCommand],
                           str,
                           typing.Sequence[str]]


def get_pager_command(pager_command: CommandType) -> PagerCommand:
    if isinstance(pager_command, PagerCommand):
        return pager_command

    if isinstance(pager_command, str):
        return CustomPager(pager_command)

    if isinstance(pager_command, collections.abc.Sequence):
        return UserSpecifiedPager(*pager_command)

    if callable(pager_command):
        cmd = pager_command()
        if isinstance(cmd, PagerCommand):
            return cmd

    raise TypeError('Invalid type for ``pager_command``')
